import attr


@attr.s
class Result:
	"""
	Base results object that holds the number of right and wrong notes.

	Arg
		right (int): The number of correct Mupix Object Elements
		wrong (int): The number of wrong Mupix Object Elements
		name (str): The name this results object
	"""
	right = attr.ib(init=False, type=int, default=0)
	wrong = attr.ib(init=False, type=int, default=0)

	name = attr.ib(type=str)
	@name.default
	def _get_name(self):
		"""
		Get the name of this Results object, even when it's inherited so its still available
		when calling an attrs as a dictionary.
		"""
		return str(self.__class__.__name__)

	def asdict(self):
		"""
		This is a just a method to unify. I'd rather have a method to turn a class object
		into a dictionary, instead of how attrs works like this.
		"""
		return attr.asdict(self)


@attr.s
class NoteStepResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Pitches.
	"""


@attr.s
class NoteNameResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Pitches.
	"""


@attr.s
class NoteDurationResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Durations.
	"""


@attr.s
class NoteOctaveResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Octaves.
	"""


@attr.s
class NoteAccidentalResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Accidentals
	"""


@attr.s
class NoteArticulationResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Articulations
	"""


@attr.s
class NoteStemDirectionResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Stem Directions
	"""


@attr.s
class NoteBeamResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Beams
	"""


@attr.s
class NoteVoiceResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Voices
	"""


@attr.s
class NoteTieStyle(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Tie Style
	"""


@attr.s
class NoteTieType(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Tie Type
	"""


@attr.s
class NoteTiePlacement(Result):
	"""
	Results Object that holds the number of correct and incorrect Note Tie Placement
	"""


@attr.s
class NoteTotalResult(Result):
	"""
	The accumulated total of all the Note related Results Objects
	"""


@attr.s
class RestArticulationResult(Result):
	"""
	Articulations can not exist by themselves in an empty measure. They must be attached to a note
	or a rest.
	"""


@attr.s
class RestDurationResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Rest Durations
	"""


@attr.s
class RestVoiceResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Rest Voices
	"""


@attr.s
class RestTotalResult(Result):
	"""
	The accumulated total of all the Rest related Results Objects
	"""


@attr.s
class TimeSignatureNumeratorResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Time Signature Numerator values
	"""


@attr.s
class TimeSignatureDenominatorResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Time Signature Denominator values
	"""


@attr.s
class TimeSignatureTotalResult(Result):
	"""
	The accumulated total of all the Time Signature related Results Objects
	"""


@attr.s
class KeySignatureStepResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Key Signature
	steps.
	"""


@attr.s
class KeySignatureModeResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Key Signature
	modes.
	"""


@attr.s
class KeySignatureOnsetResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Key Signature
	onsets.
	"""


@attr.s
class KeySignatureTotalResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Key Signature
	properties.
	"""


@attr.s
class ClefNameResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Clef names.
	"""


@attr.s
class ClefLineResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Clef lines.
	"""


@attr.s
class ClefOctaveResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Clef octaves.
	"""


@attr.s
class ClefOnsetResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Clef onsets.
	"""


@attr.s
class ClefTotalResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Clef properties.
	"""


@attr.s
class SpannerNameResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Spanner names.
	"""


@attr.s
class SpannerPlacementResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Spanner placement.
	"""


@attr.s
class SpannerLengthResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Spanner length.
	"""


@attr.s
class SpannerTotalResult(Result):
	"""
	Results Object that holds the number of correct and incorrect Spanner properties.
	"""


@attr.s
class DynamicOnsetResult(Result):
	"""
	"""


@attr.s
class DynamicNameResult(Result):
	"""
	"""


@attr.s
class DynamicTotalResult(Result):
	"""
	"""



import attr
import numpy


@attr.s
class Matrix:
	"""Matrix object for Sequence Alignment
	"""
	true_data_len = attr.ib(type=int)
	test_data_len = attr.ib(type=int)

	def __attrs_post_init__(self):
		"""
		Creating new values here will be hidden from view when printing the output.
		"""
		self.m = numpy.zeros((self.true_data_len, self.test_data_len))
		self.x = numpy.zeros((self.true_data_len, self.test_data_len))
		self.y = numpy.zeros((self.true_data_len, self.test_data_len))

		self.pointer = numpy.zeros((self.true_data_len, self.test_data_len))
		self.x_pointer = numpy.zeros((self.true_data_len, self.test_data_len))
		self.y_pointer = numpy.zeros((self.true_data_len, self.test_data_len))


@attr.s
class SequenceAlignment:
	"""Sequence alignment abstraction class
	@timothydereuse 2019
	https://github.com/DDMAL/text-alignment/blob/ba44cb090caf63514cfe925abcc3afd1f2585a1d/textSeqCompare.py#L1-L191
	"""
	true_data = attr.ib()
	test_data = attr.ib()
	aligned_true_data = attr.ib(init=False, default=[])
	aligned_test_data = attr.ib(init=False, default=[])

	match = attr.ib(kw_only=True, default=10)
	mismatch = attr.ib(kw_only=True, default=-5)
	gap_extend = attr.ib(kw_only=True, default=-1)

	gap_open_x = attr.ib(kw_only=True, default=-10)
	gap_extend_x = attr.ib(kw_only=True, default=-1)
	gap_open_y = attr.ib(kw_only=True, default=-10)
	gap_extend_y = attr.ib(kw_only=True, default=-1)

	_skips = attr.ib(init=False, default=0)

	def __attrs_post_init__(self):
		# Convert to list
		self.true_data = self._type_check(self.true_data)
		self.test_data = self._type_check(self.test_data)

		# Add extra character at the end of both, and base the alignment on that.
		self.true_data = self.true_data + [" "]
		self.test_data = self.test_data + [" "]

		self.matrix = Matrix(
			len(self.true_data),
			len(self.test_data),
		)

		self.populate()

	def __str__(self):
		# return f"""{{\n  true_data: "{"".join(self.true_data)}"\n  test_data: "{"".join(self.test_data)}"\n  aligned_true_data: "{"".join(self.aligned_true_data)}"\n  aligned_test_data: "{"".join(self.aligned_test_data)}"\n}}"""
		return """{{\n  true_data: "{}"\n  test_data: "{}"\n  aligned_true_data: "{}"\n  aligned_test_data: "{}"\n}}""".format(  # noqa
			"".join(map(str, self.true_data)),
			"".join(map(str, self.test_data)),
			"".join(map(str, self.aligned_true_data)),
			"".join(map(str, self.aligned_test_data)),
		)

	def _type_check(self, input):
		if isinstance(input, str):
			input = [char for char in input]
		elif isinstance(input, list):
			pass
		else:
			raise Exception(
				f"{self.__class__}\n\n" +
				f"Input data of type {type(input)} is not acceptable.\n" +
				"Please use a String or a List instead."
			)
		return input

	def scoring_method(self, true, test):
		raise Exception(
			f"{self.__class__}\n\n" +
			f"You forgot to specify a scoring_method() method."
		)

	def populate(self):
		raise Exception(
				f"{self.__class__}\n\n" +
				f"You forgot to specify a populate() method."
			)

	def find_shortest_path(self):
		"""
		Fill the array with aligned elements, true skip ("_") or test skip ("_").
		"""
		while(self.xpt > 0 and self.ypt > 0):
			# If both values true/test are the same.
			if self.mpt == 0:
				self.true_align.append(self.true_data[self.xpt - 1])
				self.test_align.append(self.test_data[self.ypt - 1])

				self.xpt -= 1
				self.ypt -= 1

			elif self.mpt == 1:
				self.true_align.append(self.true_data[self.xpt - 1])
				self.test_align.append("_")
				self._skips += 1
				self.xpt -= 1

			elif self.mpt == 2:
				self.true_align.append("_")
				self.test_align.append(self.test_data[self.ypt - 1])
				self._skips += 1
				self.ypt -= 1

			self.mpt = self.matrix.pointer[self.xpt][self.ypt]

		while self.ypt > 0:
			self.true_align.append("_")
			self.test_align.append(self.test_data[self.ypt - 1])
			self.ypt -= 1
		while self.xpt > 0:
			self.true_align.append(self.true_data[self.xpt - 1])
			self.test_align.append("_")
			self.xpt -= 1

		self.aligned_true_data = reversed(self.true_align[1:])
		self.aligned_test_data = reversed(self.test_align[1:])

	def traceback(self):
		self.true_align = []
		self.test_align = []
		self.align_record = []
		self.pointer_record = []
		self.xpt = len(self.true_data) - 1
		self.ypt = len(self.test_data) - 1
		self.mpt = self.matrix.pointer[self.xpt][self.ypt]

		# Align the extra characters at the end.
		self.true_align += self.true_data[self.xpt]
		self.test_align += self.test_data[self.ypt]

		return self.find_shortest_path()


@attr.s
class AffineNeedlemanWunsch(SequenceAlignment):
	"""Basic Needleman-Wunsch abstraction class.

	As you can see, you just need to overwrite the `scoring_method` and define
	how to populate your matrix by overwriting the `populate` method. Once both
	these methods are defined, make sure you call `self.traceback()` at the end
	of your `populate` method to create your fully-formed matrix and give output
	the algorithms best guess.

	Align each note object with each other. Understandably this is not the best, but it
		serves more as a proof-of-concept or a placeholder to be built later.

		- Notes           are aligned by step names
		- Restes          are aligned by measure number as a single char
		- TimeSignatures  are aligned by measure number as a single char
		- KeySignatures   are aligned by measure number as a single char
		- Clefs           are aligned by measure number as a single char

	:param [true_data]: A character string that represents what needs to be aligned,
		or a list of objects, this is where you put your ground truth.
	:type [true_data]: String, List

	:param [test_data]: A character string that represents what needs to be aligned,
		or a list of objects, this is where you put your ground truth.
	:type [true_data]: String, List

	:return: Returns a json object with 4 keys (both original data, and both aligned data)
	:rtype: JSON
	"""
	def scoring_method(self, true, test):
		"""
		Grading method for comparing the two elements in scope.
		"""
		if true == test:
			return 1
		else:
			return 0

	def populate(self):
		for index, value in enumerate(self.true_data):
			self.matrix.m[index][0] = self.gap_extend * index
			self.matrix.x[index][0] = float("-inf")
			self.matrix.y[index][0] = float("-inf")
		for index, value in enumerate(self.test_data):
			self.matrix.m[0][index] = self.gap_extend * index
			self.matrix.x[0][index] = self.gap_extend * index
			self.matrix.y[0][index] = float("-inf")

		for i in range(1, len(self.true_data)):
			for j in range(1, len(self.test_data)):
				eq_res = self.scoring_method(self.true_data[i - 1], self.test_data[j - 1])
				match_score = self.match if eq_res else self.mismatch

				matrix_values = [
					self.matrix.m[i - 1][j - 1],
					self.matrix.x[i - 1][j - 1],
					self.matrix.y[i - 1][j - 1],
				]
				self.matrix.m[i][j] = max(matrix_values) + match_score
				self.matrix.pointer[i][j] = matrix_values.index(max(matrix_values))

				x_matrix_values = [
					self.matrix.m[i - 1][j] + self.gap_open_x + self.gap_extend_x,
					self.matrix.x[i - 1][j] + self.gap_extend_x,
					self.matrix.y[i - 1][j] + self.gap_open_x + self.gap_extend_x,
				]
				self.matrix.x[i][j] = max(x_matrix_values)
				self.matrix.x_pointer[i][j] = x_matrix_values.index(max(x_matrix_values))

				y_matrix_values = [
					self.matrix.m[i][j - 1] + self.gap_open_y + self.gap_extend_y,
					self.matrix.x[i][j - 1] + self.gap_open_y + self.gap_extend_y,
					self.matrix.y[i][j - 1] + self.gap_extend_y,
				]
				self.matrix.y[i][j] = max(y_matrix_values)
				self.matrix.y_pointer[i][j] = y_matrix_values.index(max(y_matrix_values))

		self.traceback()


@attr.s
class AdvancedAffineNeedlemanWunsch(AffineNeedlemanWunsch):
	def scoring_method(self, true, test):
		"""
		Grading method for comparing the two elements in scope.

			- Values are randomly chosen
			- Eventually create a genetic algorithm to find optimal values
		"""
		score = 0

		try:
			true.asname()
			test.asname()
		except AttributeError:
			return score

		# If both objects are Notes
		if true.asname() == "Note" and test.asname() == "Note":
			score += 1 if true.octave == test.octave else -1
			# score += 1 if true.voice == test.voice else -1
			score += 4 if true.step == test.step else -4
			score += 2 if true.duration == test.duration else -2
			# score += 1 if true.beam == test.beam else -1
			score += 1 if true.accidental == test.accidental else -1
			# score += 1 if true.stemdirection == test.stemdirection else -1

		if true.asname() == "Rest" and test.asname() == "Rest":
			score += 1 if true.voice == test.voice else -1
			score += 5 if true.duration == test.duration else -5

		if true.asname() == "TimeSignature" and test.asname() == "TimeSignature":
			score += 2 if true.numerator == test.numerator else -2
			score += 2 if true.denominator == test.denominator else -2

		if true.asname() == "KeySignature" and test.asname() == "KeySignature":
			score += 2 if true.step == test.step else -2
			score += 2 if true.mode == test.mode else -2
			score += 2 if true.onset == test.onset else -2

		if true.asname() == "Clef" and test.asname() == "Clef":
			score += 5 if true.name == test.name else -5
			score += 2 if true.line == test.line else -2
			score += 2 if true.octave == test.octave else -2
			score += 2 if true.onset == test.onset else -2

		if true.asname() == "Spanner" and test.asname() == "Spanner":
			score += 5 if true.name == test.name else -5
			score += 2 if true.placement == test.placement else -2
			score += 1 if true.length == test.length else -1

		# If the objects are the same type
		score += 10 if true.part == test.part else -10
		score += 5 if true.measure == test.measure else -5
		score += 5 if true.onset == test.onset else -5
		score += 5 if true.asname() == test.asname() else -5

		return score

	def populate(self):
		for index, value in enumerate(self.true_data):
			self.matrix.m[index][0] = self.gap_extend * index
			self.matrix.x[index][0] = float("-inf")
			self.matrix.y[index][0] = float("-inf")
		for index, value in enumerate(self.test_data):
			self.matrix.m[0][index] = self.gap_extend * index
			self.matrix.x[0][index] = self.gap_extend * index
			self.matrix.y[0][index] = float("-inf")

		for i, _ in enumerate(self.true_data):
			for j, _ in enumerate(self.test_data):
				match_score = self.scoring_method(self.true_data[i - 1], self.test_data[j - 1])

				matrix_values = [
					self.matrix.m[i - 1][j - 1],
					self.matrix.x[i - 1][j - 1],
					self.matrix.y[i - 1][j - 1],
				]
				self.matrix.m[i][j] = max(matrix_values) + match_score
				self.matrix.pointer[i][j] = matrix_values.index(max(matrix_values))

				x_matrix_values = [
					self.matrix.m[i - 1][j] + self.gap_open_x + self.gap_extend_x,
					self.matrix.x[i - 1][j] + self.gap_extend_x,
					self.matrix.y[i - 1][j] + self.gap_open_x + self.gap_extend_x,
				]
				self.matrix.x[i][j] = max(x_matrix_values)
				self.matrix.x_pointer[i][j] = x_matrix_values.index(max(x_matrix_values))

				y_matrix_values = [
					self.matrix.m[i][j - 1] + self.gap_open_y + self.gap_extend_y,
					self.matrix.x[i][j - 1] + self.gap_open_y + self.gap_extend_y,
					self.matrix.y[i][j - 1] + self.gap_extend_y,
				]
				self.matrix.y[i][j] = max(y_matrix_values)
				self.matrix.y_pointer[i][j] = y_matrix_values.index(max(y_matrix_values))

		self.traceback()


if __name__ == "__main__":
	seq1 = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit '
	seq2 = 'LoLorem fipsudor ..... st emet, c.nnr adizcing eelilit'

	print(
		AffineNeedlemanWunsch(
			# # Accepts Lists
			# [char for char in seq1], [char for char in seq2],

			# # Accepts Strings
			seq1, seq2,

			# # Raises and Exception on everything else.

			# # Tuple
			# (1,2,3,4), seq2,

			# # Int
			# 1243, seq1,

			# # Float
			# 1.123, seq1,

			# # Dicts
			# {"asdf": "zxcv"}, seq2,
		)
	)



"""
The properties of these objects are named in a specific way for a reason. Any
property that is not to be evaluated has an underscore in them
(eg: _music21_object). If the property requires more than a word to describe
it, then write it in camelCase style.
"""
import attr


@attr.s
class Marking:
	"""
	The base class for all the components of a Mupix object

	:property [_music21_object]: The original music21 object, in case it is needed.

	:property [part]: An integer representing the instrument (1 for the first instrument, etc.)

	:property [measure]: The measure in which this object can be found. It is
	inferred by the music21 object.

	:property [onset]: The time in the measure where this event occurs. It is
	inferred by the music21 object.
	"""
	_music21_object = attr.ib(eq=False)
	part = attr.ib(type=int)

	measure = attr.ib(init=False)
	@measure.default
	def _get_measure(self):
		return int(self._music21_object.measureNumber)

	onset = attr.ib(init=False, type=str)
	@onset.default
	def _get_onset(self):
		return str(self._music21_object.offset)

	def asdict(self):
		"""
		Return the object as a JSON serializable python dictionary.
		"""
		tmp = attr.asdict(self)
		del tmp["_music21_object"]
		return tmp

	def asname(self):
		"""
		Returns the name of the current class (or subclass)
		"""
		string = str(self.__class__).split(".")[-1].replace("Object", "")
		return string[0].lower() + string[1:-2] + "s"


@attr.s
class MusicalEvent(Marking):
	"""
	A musical event is an **active** marking in a score, rather than **passive**
	marking like a key signature or a time signature. Active markings are
	effective when read and discarded when finished.

	:property [duration]: The amount of time the event will last.

	:property [voice]: In the case of multiple instruments within a staff, this
	is a numerical representation of which instrument it represents.

	:property [articulation]: A musical articulation (staccato, tenuto, accents, etc.)
	"""
	duration = attr.ib(init=False, type=str, eq=False)
	@duration.default
	def _get_duration(self):
		return str(self._music21_object.quarterLength)

	voice = attr.ib(init=False, type=int)
	@voice.default
	def _get_voice(self):
		if isinstance(self._music21_object.activeSite.id, int):
			return 1
		else:
			return int(self._music21_object.activeSite.id)

	articulation = attr.ib(init=False, eq=False)
	@articulation.default
	def _get_articulation(self):
		return [item.name for item in self._music21_object.articulations]


@attr.s
class NoteObject(MusicalEvent):
	"""
	A NoteObject holds information useful to Mupix for symbolic music
	file differentiation.

	:property [step]: I believe the step in Music21 should be renamed to a
	`note.name`. A step should be similar to the concept of a scale degree,
	except that all the distances are counted in semitones relative to the key
	signature. Also note that octave differences are also removed, much like
	scale degrees.

	.. note:: For example
		In the key of C Major, if the octave information is removed

		C2 -> C8  = 0 (C, C)

		C2 -> C#8 = 1 (C, C#) that's one semitone

		C8 -> C#2 = 1 (C, C#) Still one semitone

		C2 -> G4  = 7 (C, C#, D, D#, E, F, F#, G) still 7 semitones

	:property [pitch]:
	This is what Music21 defines as a step (C, D, E, F, G, etc.)
	For continued use, grab the pitch by referencing the music21 object.

	.. warning:: This property will be deprecated in **version 0.2.0**. Use **noteObject._music21_object.step** instead.

	:property [octave]: The octave of the note (middle C is C4)

	:property [accidental]: Any note alterations of pitch (#, ##, b, bb, natural)

	:property [stemdirection]: The direction (up or down) of the note stem (if present).

	:property [beam]: How note beams are connected between adjacent notes.

	"""
	step = attr.ib(kw_only=True, eq=False, default=None)

	name = attr.ib(init=False, eq=False)
	@name.default
	def _get_name(self):
		return self._music21_object.step

	octave = attr.ib(init=False, eq=False)
	@octave.default
	def _get_octave(self):
		return self._music21_object.octave

	accidental = attr.ib(init=False, type=str, eq=False)
	@accidental.default
	def _get_accidental(self):
		note = self._music21_object
		if len(note.name) > 1:
			return note.name[1:]
		else:
			return ""

	stemdirection = attr.ib(init=False, eq=False)
	@stemdirection.default
	def _get_stem_direction(self):
		return self._music21_object.stemDirection

	beam = attr.ib(init=False, eq=False)
	@beam.default
	def _get_beam(self):
		note = self._music21_object
		note.beams.getTypes()
		# The number of "partial" elements that can appear in the beams is related to the note duration.
		# We do not want the errors be disproportionate if the duration is wrong too.
		return set([item for item in note.beams.getTypes()])

	tiestyle = attr.ib(init=False, eq=False)
	@tiestyle.default
	def _get_tie_style(self):
		"""
		"normal" or "dotted" or "dashed" or "hidden"
		"""
		try:
			return self._music21_object.tie.style
		except AttributeError:
			return None

	tietype = attr.ib(init=False, eq=False)
	@tietype.default
	def _get_tie_type(self):
		"""
		"start", "stop", "continue", "let-ring", "continue-let-ring"
		"""
		try:
			return self._music21_object.tie.type
		except AttributeError:
			return None

	tieplacement = attr.ib(init=False, eq=False)
	@tieplacement.default
	def _get_tie_placement(self):
		"""
		None, above, below
		"""
		try:
			return self._music21_object.tie.placement
		except AttributeError:
			return None


@attr.s
class RestObject(MusicalEvent):
	"""
	Same as :func:`mupix.core.MusicalEvent`
	"""
	pass


@attr.s
class TimeSignatureObject(Marking):
	"""
	This is an object that holds information related to all time signatures.
	"""
	numerator = attr.ib(init=False, eq=False)
	@numerator.default
	def _get_numerator(self):
		return self._music21_object._getNumerator()

	denominator = attr.ib(init=False, eq=False)
	@denominator.default
	def _get_denominator(self):
		return self._music21_object._getDenominator()


@attr.s
class KeySignatureObject(Marking):
	"""
	This is an object that holds information related to all key signatures.
	"""
	step = attr.ib(init=False, eq=False)
	@step.default
	def _get_step(self):
		return self._music21_object.asKey().name.split(" ")[0]

	mode = attr.ib(init=False, eq=False)
	@mode.default
	def _get_mode(self):
		return self._music21_object.asKey().name.split(" ")[1]


@attr.s
class ClefObject(Marking):
	"""
	This is an object that holds information related to all clefs.
	"""
	name = attr.ib(init=False, eq=False)
	@name.default
	def _get_name(self):
		return self._music21_object.sign

	line = attr.ib(init=False, eq=False)
	@line.default
	def _get_line(self):
		return self._music21_object.line

	octave = attr.ib(init=False, eq=False)
	@octave.default
	def _get_octave(self):
		return self._music21_object.octaveChange


@attr.s
class SpannerObject(Marking):
	"""
	This is an object that holds information related to musical objects which can span multiple
	measures like crescendos.
	"""
	name = attr.ib(init=False, eq=False)
	@name.default
	def _get_name(self):
		return self._music21_object.type

	# This is empty in Music21, need to repopulate this data.
	measure = attr.ib(init=False)
	@measure.default
	def _get_measure(self):
		return self._music21_object.getFirst().measureNumber

	placement = attr.ib(init=False, eq=False)
	@placement.default
	def _get_placement(self):
		return self._music21_object.placement

	length = attr.ib(init=False, eq=False)
	@length.default
	def _get_length(self):
		return self._music21_object.spread


@attr.s
class DynamicObject(Marking):
	"""
	This is an object that holds information related to dynamics
	"""
	name = attr.ib(init=False, eq=False)
	@name.default
	def _get_name(self):
		return self._music21_object.longName


if __name__ == "__main__":
	"""
	How to create Mupix Objects.
	"""
	from music21.stream import Score, Part, Measure
	from music21.key import KeySignature
	from music21.note import Note  # noqa

	from mupix.extra import add_step_information

	s = Score()

	p1 = Part(id="part1")
	m1 = Measure(number=1)
	m1.append(KeySignature(0))
	m1.append(Note("C4", type="eighth"))
	m2 = Measure(number=2)
	m2.append(KeySignature(1))
	m2.append(Note("G4", type="eighth"))
	p1.append([m1, m2])

	p2 = Part(id="part2")
	m1_ = Measure(number=1)
	m1_.append(KeySignature(0))
	m1_.append(Note("C4", type="eighth"))
	m2_ = Measure(number=2)
	m2_.append(KeySignature(1))
	m2_.append(Note("G4", type="eighth"))
	p2.append([m1_, m2_])

	s.append([p1, p2])

	notes = [NoteObject(item, 1) for item in s.recurse().notes if not item.isChord]
	keySignatures = [KeySignatureObject(item, 1) for item in s.recurse().getElementsByClass("KeySignature")]  # noqa

	print(notes)
	notes = add_step_information(notes, keySignatures)
	print(notes)



"""
//FIXME: Move from RE to lxml library only when the following three conditions are met
	//TODO: Get all major commercial music notation software (Missing one)
	//TODO: Create testing sheets using all notation software
	//TODO: Create an automated method for testing multiple versions of Notation software (Almost done)
"""

import re
import operator

import attr
import music21

from mupix.core import (
	NoteObject,
	RestObject,
	TimeSignatureObject,
	KeySignatureObject,
	ClefObject,
	SpannerObject,
	DynamicObject,
)
from mupix.result_objects import (
	Result,
	NoteNameResult,
	NoteStepResult,
	NoteDurationResult,
	NoteOctaveResult,
	NoteAccidentalResult,
	NoteArticulationResult,
	NoteStemDirectionResult,
	NoteBeamResult,
	NoteVoiceResult,
	NoteTieStyle,
	NoteTieType,
	NoteTiePlacement,
	NoteTotalResult,
	RestArticulationResult,
	RestDurationResult,
	RestVoiceResult,
	RestTotalResult,
	TimeSignatureNumeratorResult,
	TimeSignatureDenominatorResult,
	TimeSignatureTotalResult,
	KeySignatureStepResult,
	KeySignatureModeResult,
	KeySignatureOnsetResult,
	KeySignatureTotalResult,
	ClefNameResult,
	ClefLineResult,
	ClefOctaveResult,
	ClefOnsetResult,
	ClefTotalResult,
	SpannerNameResult,
	SpannerPlacementResult,
	SpannerLengthResult,
	SpannerTotalResult,
	DynamicOnsetResult,
	DynamicNameResult,
	DynamicTotalResult,
)
from mupix.extra import (
	add_step_information,
	normalize_object_list,
	return_char_except,
	boundary_search,
)


@attr.s
class MupixObject():
	"""A MupixObject holds information for an entire score.
	A Mupix Object contains multiple lists of the contents within a symbolic music-file.

	:param [notes]: A list of objects related to notes
	:type [notes]: List or Result object

	:param [rests]: A list of objects related to rests
	:type [rests]: List or Result object

	:param [timeSignatures]: A list of objects related to time signatures
	:type [timeSignatures]: List or Result object

	:param [keySignatures]: A list of objects related to key signatures
	:type [keySignatures]: List or Result object

	:param [clefs]: A list of objects related to clefs
	:type [clefs]: List or Result object

	:param [parts]: A number representing the number of total staffs per system
	:type [parts]: Integer

	:param [error_description]: More detailed error information
	:type [error_description]: Dictionary
	"""
	notes = attr.ib(kw_only=True,)
	rests = attr.ib(kw_only=True,)
	timeSignatures = attr.ib(kw_only=True,)
	keySignatures = attr.ib(kw_only=True,)
	clefs = attr.ib(kw_only=True,)
	spanners = attr.ib(kw_only=True,)
	dynamics = attr.ib(kw_only=True,)
	parts = attr.ib(kw_only=True, type=int, validator=[attr.validators.instance_of(int)])
	error_description = attr.ib(kw_only=True, type=dict, validator=[attr.validators.instance_of(dict)])
	software_vendor = attr.ib(kw_only=True, type=list, default=[], validator=[attr.validators.instance_of(list)])

	@notes.validator
	@rests.validator
	@timeSignatures.validator
	@keySignatures.validator
	@clefs.validator
	@spanners.validator
	@dynamics.validator
	def check(self, attribute, value):
		"""
		Type-check for notes, rests, time signatures, key signatures, and, clefs.
		They should be either a list, or a Result object for eventual output.
		"""
		if not isinstance(value, list) and not isinstance(value, Result):
			raise ValueError(f"Must be a list or Results Object. {type(value)}")

	def ret(self):
		"""
		Return all information about the Mupix object as a tuple.
		"""
		# return self.notes, self.rests, self.timeSignatures, self.keySignatures, self.clefs, self.error_description
		return self.notes, self.rests, self.timeSignatures, self.keySignatures, self.clefs, self.spanners, self.dynamics, self.error_description

	def __iter__(self):
		return iter(self.ret())

	@classmethod
	def from_filepath(cls, filepath):
		"""
		.. note::

			Using activesite from music21 won't work because the key
			signature would be in measure one, and in measure two there is no
			information about key anymore. So :func:`mupix.base.add_step_information`
			was written.

			.. code-block:: python

			notes += [NoteObject(
				item,           # Music21 Object
				parts_index,    # Part number
				step=Interval(  # A step defined in the NoteObjects from mupix/base.py
				noteStart=item._getActiveSite().keySignature.asKey().tonic,
				noteEnd=item).semitones % 12) for item in parts.recurse().notes if not item.isChord]


		:param [filepath]: A character string that represents the filepath and filename of the file to open.
		:type [filepath]: String

		:return: A fully populated Mupix Object, with all the components of the symbolic music file analysized and sorted in their sections.
		:rtype: Mupix Object
		"""

		with open(filepath, "r") as f:
			data = f.read()

		software_vendor = re.finditer(r"(?<=<software>).+(?=<\/software>)", data).__next__().group().split(" ")

		notes, rests, timeSignatures, keySignatures, clefs, spanners, dynamics = [], [], [], [], [], [], []
		_keySignatures = []
		# _clefs = []
		for parts_index, parts in enumerate(music21.converter.parseFile(filepath, forceSource=True).recurse().getElementsByClass("Part"), 1):  # noqa
			notes += [NoteObject(item, parts_index) for item in parts.recurse().notes if not item.isChord]
			rests += [RestObject(item, parts_index) for item in parts.recurse().notesAndRests if not item.isNote]
			timeSignatures += [TimeSignatureObject(item, parts_index) for item in parts.recurse().getElementsByClass("TimeSignature")]  # noqa
			keySignatures += [KeySignatureObject(item, parts_index) for item in parts.recurse().getElementsByClass("KeySignature")]  # noqa
			clefs += [ClefObject(item, parts_index) for item in parts.recurse().getElementsByClass("Clef")]
			dynamics += [DynamicObject(item, parts_index) for item in parts.recurse().getElementsByClass("Dynamic")]

			for measure in parts.recurse().getElementsByClass("Measure"):
				spanners += [SpannerObject(item, parts_index) for item in measure.activeSite.spanners]

		########################################################
		# Expand keys
		#
		# Because even with activeState, note objects can't find 
		# the key for some reason?
			_keySignatures += [KeySignatureObject(item, parts_index) for item in parts.recurse().getElementsByClass("KeySignature")]  # noqa
			# _clefs += [ClefObject(item, parts_index) for item in parts.recurse().getElementsByClass("Clef")]

		try:
			measuresInScore = max(notes + rests, key=operator.attrgetter('measure')).measure
		except ValueError:
			measuresInScore = 0
			parts_index = 0

		_keySignatures = normalize_object_list(object_list=_keySignatures, total_measures=measuresInScore, total_parts=parts_index)
		# _clefs = normalize_object_list(object_list=_clefs, total_measures=measuresInScore, total_parts=parts_index)

		# only once _keySignatures are normalized can we add the step information.
		notes = add_step_information(notes, _keySignatures)
		########################################################

		return cls(
			notes=notes,
			rests=rests,
			timeSignatures=timeSignatures,
			keySignatures=keySignatures,
			clefs=clefs,
			parts=parts_index,
			spanners=spanners,
			dynamics=dynamics,
			error_description={},
			software_vendor=software_vendor,
		)


class BaseCompareClass(MupixObject):
	"""
	The base comparison class for Mupix Objects.
	"""
	def __init__(self, true_filepath: str, test_filepath: str, do_not_count: list = []):
		# for result_to_exclude in do_not_count:
		#   del self.__getattribute__(result_to_exclude)

		# Notes
		self.notes = []
		self.notes_step = NoteStepResult()
		self.notes_name = NoteNameResult()
		self.notes_duration = NoteDurationResult()
		self.notes_octave = NoteOctaveResult()
		self.notes_accidental = NoteAccidentalResult()
		self.notes_articulation = NoteArticulationResult()
		self.notes_stemdirection = NoteStemDirectionResult()
		self.notes_beam = NoteBeamResult()
		self.notes_voice = NoteVoiceResult()
		self.notes_tiestyle = NoteTieStyle()
		self.notes_tietype = NoteTieType()
		self.notes_tieplacement = NoteTiePlacement()
		self.notes_total = NoteTotalResult()

		# Rests
		self.rests = []
		self.rests_articulation = RestArticulationResult()
		self.rests_duration = RestDurationResult()
		self.rests_voice = RestVoiceResult()
		self.rests_total = RestTotalResult()

		# Time Signatures
		self.timeSignatures = []
		self.timeSignatures_numerator = TimeSignatureNumeratorResult()
		self.timeSignatures_denominator = TimeSignatureDenominatorResult()
		self.timeSignatures_total = TimeSignatureTotalResult()

		# Key Signatures
		self.keySignatures = []
		self.keySignatures_step = KeySignatureStepResult()
		self.keySignatures_mode = KeySignatureModeResult()
		self.keySignatures_onset = KeySignatureOnsetResult()
		self.keySignatures_total = KeySignatureTotalResult()

		# Clefs
		self.clefs = []
		self.clefs_name = ClefNameResult()
		self.clefs_line = ClefLineResult()
		self.clefs_octave = ClefOctaveResult()
		self.clefs_onset = ClefOnsetResult()
		self.clefs_total = ClefTotalResult()

		# Spanners
		self.spanners = []
		self.spanners_name = SpannerNameResult()
		self.spanners_placement = SpannerPlacementResult()
		self.spanners_length = SpannerLengthResult()
		self.spanners_total = SpannerTotalResult()

		# Dynamics
		self.dynamics = []
		self.dynamics_onset = DynamicOnsetResult()
		self.dynamics_name = DynamicNameResult()
		self.dynamics_total = DynamicTotalResult()


		self.error_description = {}

		# Parse both files
		self.true_data = MupixObject.from_filepath(true_filepath)
		self.test_data = MupixObject.from_filepath(test_filepath)

	def _return_object_names(self):
		"""
		Returns all the objects

		For example:
			['clefs', 'keySignatures', 'notes', 'rests', 'timeSignatures', 'spanners']
		"""
		return [item for item in dir(self) if "_" not in item and item not in ["check", "ret"]]

	def _return_parameter_names(self, field):
		"""
		For a specific object, return all items

		For notes:
			['notes_accidental', 'notes_articulation', 'notes_beam', 'notes_duration', 'notes_octave', 'notes_step', 'notes_stemdirection']
		"""
		return [item for item in dir(self) if field in item and "_" in item and "total" not in item]

	def _compare_expand_objects_same(self, true_object, test_object, object_):
		"""
		Compare the properties of two same objects for any property that may be incorrect.
		"""
		for result_parameter in self._return_parameter_names(object_):
			property_ = result_parameter.split("_")[-1]
			try:
				if true_object.__getattribute__(property_) == test_object.__getattribute__(property_):
					self.__getattribute__(result_parameter).right += 1
				else:
					out = (
						f"{true_object.part}-{true_object.measure}-{true_object.onset}=>{test_object.part}-"
						f"{test_object.measure}-{test_object.onset}__{true_object.__getattribute__(property_)}"
						f"_{test_object.__getattribute__(property_)}"
					)
					try:
						self.error_description[result_parameter].append(out)
					except KeyError:
						self.error_description[result_parameter] = [out]
					self.__getattribute__(result_parameter).wrong += 1
			except AttributeError:
				raise Exception(type(true_object), type(test_object), "What happened here???")

	def _compare_expand_objects_different(self, true_object, test_object, object_):
		for parameter in self._return_parameter_names(object_):

			if isinstance(test_object, str):
				out = (
					f"{true_object.part}-{true_object.measure}-{true_object.onset}=>"
					"skip adjustment"
				)
			elif isinstance(true_object, str):
				out = (
					f"skip adjustment=>{test_object.part}"
					f"-{test_object.measure}-{test_object.onset}__{parameter}"
				)
			else:
				out = (
					f"{true_object.part}-{true_object.measure}-{true_object.onset}=>{test_object.part}"
					f"-{test_object.measure}-{test_object.onset}__{parameter}"
				)
			try:
				self.error_description[parameter].append(out)
			except KeyError:
				self.error_description[parameter] = [out]

			self.__getattribute__(parameter).wrong += 1

	def _compare(self, true_object, test_object):
		"""
		Compare two Mupix Objects.

			- If the sequence alignment believes there is an extra note in the OMR output, simply deleting the extra note
			should be the solution. Removing each wrong element individually and finally removing the note is not how a
			human would fix this type of error.

			- Then check for each parameter available in the type of marking, count an error for each wrong element and a
			right for every correct. This function should works for all Mupix objects.
		"""
		# If object is misaligned
		try:
			self.true_part = true_object.part
			self.true_onset = true_object.onset
			self.true_mu21_obj = true_object._music21_object
		except AttributeError:
			self.true_part = "_"
			self.true_onset = ""
			self.true_mu21_obj = ""

		# If object is missaligned
		try:
			self.test_part = test_object.part
			self.test_onset = test_object.onset
			self.test_mu21_obj = test_object._music21_object
		except AttributeError:
			self.test_part = "_"
			self.test_onset = ""
			self.test_mu21_obj = ""

		# As much as I liked how small and compact this was vs writing them out,
		# it became hard to find out where the real problem was...
		#
		# # If there is an extra object in the test data, it's better to just delete
		# # it and note it as wrong for being additional incorrect data.
		# if true_object == "_":
		#   self.__getattribute__(f"{test_object.asname()}_total").wrong += 1
		#   self.error_description[f"part{self.true_part}_onset{self.true_onset}_not_aligned_{self.true_mu21_obj}"] = f"part{self.test_part}_onset{self.test_onset}_{self.test_mu21_obj}"  # noqa
		# else:
		#   for param in self._return_parameter_names(true_object.asname()):
		#     try:
		#       if true_object.__getattribute__(param.split("_")[-1]) == test_object.__getattribute__(param.split("_")[-1]):
		#         self.__getattribute__(param).right += 1
		#       else:
		#         self.__getattribute__(param).wrong += 1
		#         self.error_description[f"part{self.true_part}_onset{self.true_onset}_{param}_{self.true_mu21_obj}"] = f"part{self.test_part}_onset{self.test_onset}_{self.test_mu21_obj}"  # noqa
		#     except AttributeError:
		#       # If the object needs to be created, each parameter needs to be added individually, +1 error each.
		#       self.__getattribute__(param).wrong += 1
		#       self.error_description[f"part{self.true_part}_onset{self.true_onset}_{param}_{self.true_mu21_obj}"] = f"part{self.test_part}_onset{self.test_onset}_{self.test_mu21_obj}"  # noqa

		if isinstance(true_object, NoteObject) and isinstance(test_object, NoteObject):
			self._compare_expand_objects_same(true_object, test_object, "notes")

			for i in true_object.articulation:
				if i in test_object.articulation:
					self.notes_articulation.right += 1
				else:
					out = (
						f"{true_object.part}-{true_object.measure}-{true_object.onset}=>{test_object.part}"
						f"-{test_object.measure}-{test_object.onset}"
					)
					try:
						self.error_description["note_articulation"].append(out)
					except KeyError:
						self.error_description["note_articulation"] = [out]
					self.notes_articulation.wrong += 1

		elif (isinstance(true_object, NoteObject) and isinstance(test_object, str)) or (isinstance(true_object, str) and isinstance(test_object, NoteObject)):
			self._compare_expand_objects_different(true_object, test_object, "notes")

		elif isinstance(true_object, RestObject) and isinstance(test_object, RestObject):
			self._compare_expand_objects_same(true_object, test_object, "rests")

			for i in true_object.articulation:
				if i in test_object.articulation:
					self.rests_articulation.right += 1
				else:
					out = (
						f"{true_object.part}-{true_object.measure}-{true_object.onset}=>{test_object.part}-"
						f"{test_object.measure}-{test_object.onset}"
					)
					try:
						self.error_description["rest_articulation"].append(out)
					except KeyError:
						self.error_description["rest_articulation"] = [out]
					self.rests_articulation.wrong += 1

		elif (isinstance(true_object, RestObject) and isinstance(test_object, str)) or (isinstance(true_object, str) and isinstance(test_object, RestObject)):
			self._compare_expand_objects_different(true_object, test_object, "rests")

		elif isinstance(true_object, TimeSignatureObject) and isinstance(test_object, TimeSignatureObject):
			self._compare_expand_objects_same(true_object, test_object, "timeSignatures")

		elif (isinstance(true_object, TimeSignatureObject) and isinstance(test_object, str)) or (isinstance(true_object, str) and isinstance(test_object, TimeSignatureObject)):
			self._compare_expand_objects_different(true_object, test_object, "timeSignatures")

		elif isinstance(true_object, KeySignatureObject) and isinstance(test_object, KeySignatureObject):
			self._compare_expand_objects_same(true_object, test_object, "keySignatures")

		elif (isinstance(true_object, KeySignatureObject) and isinstance(test_object, str)) or (isinstance(true_object, str) and isinstance(test_object, KeySignatureObject)):
			self._compare_expand_objects_different(true_object, test_object, "keySignatures")

		elif isinstance(true_object, ClefObject) and isinstance(test_object, ClefObject):
			self._compare_expand_objects_same(true_object, test_object, "clefs")

		elif (isinstance(true_object, ClefObject) and isinstance(test_object, str)) or (isinstance(true_object, str) and isinstance(test_object, ClefObject)):
			self._compare_expand_objects_different(true_object, test_object, "clefs")

		elif isinstance(true_object, SpannerObject) and isinstance(test_object, SpannerObject):
			self._compare_expand_objects_same(true_object, test_object, "spanners")

		elif (isinstance(true_object, SpannerObject) and isinstance(test_object, str)) or (isinstance(true_object, str) and isinstance(test_object, SpannerObject)):
			self._compare_expand_objects_different(true_object, test_object, "spanners")

	def _object_split(self):
		"""
		Align Objects together by comparing voice, measure and onset.
		"""
		for obj in self._return_object_names():
			for true_object in self.true_data.__getattribute__(obj):
				for test_object in self.test_data.__getattribute__(obj):
					if true_object == test_object:
						self._compare(true_object, test_object)

	def _total(self):
		"""
		Automatically add the number of right/wrong attributes to the total of each
		param (notes, rests).

		This function also checks if there are less errors by fixing the key with
		all the pitches, or most the pitches are correct. But it is not smart,
		it takes a global average of all the instruments together and not based on
		the part.
		"""

		try:
			if self.notes_step.wrong > self.notes_name.wrong:
				del self.notes_step
			else:
				del self.notes_name
		except AttributeError:
			self.error_description["Error"] = "Could not remove note_step or note_name"

		for obj in self._return_object_names():
			for params in self._return_parameter_names(obj):
				# Add the detailed results to the list of objects
				self.__getattribute__(obj).append(self.__getattribute__(params))
				self.__getattribute__(f"{obj}_total").right += self.__getattribute__(params).right
				self.__getattribute__(f"{obj}_total").wrong += self.__getattribute__(params).wrong

			self.__getattribute__(obj).append(self.__getattribute__(f"{obj}_total"))

	def _rebuild(self, aligned_data, unaligned_data):
		"""
		Rebuild the list of items according to how the AffineNeedlemanWunsch aligned them.
		"""
		output = []
		track = 0
		for entry in aligned_data:
			if entry != "_" and entry != " ":
				output.append(unaligned_data[track])
				track += 1
			else:
				output.append("_")

		return output

	def basic_sequence_alignment(self, func):
		"""
		This will align all the objects based on the sequence alignment class that
		is inserted as the function argument. The problem is that it uses specific
		parameters of each object to align them. It would be much better to align
		using values for each parameter.

		.. note:: Align each note object with each other. Understandably this is not
			the best, but it serves more as a proof-of-concept or a placeholder to be
			built later.

			- Notes           are aligned by step names

			- Restes          are aligned by measure number as a single char

			- TimeSignatures  are aligned by measure number as a single char

			- KeySignatures   are aligned by measure number as a single char

			- Clefs           are aligned by measure number as a single char

			- Spanners 				are aligned by spanner names

		:param [func]: A class that inherited from the SequenceAlignment class
			which is defined in the sequence_alignment.py file.
		:type [func]: SequenceAlignment
		"""

		# Notes
		true_notes = [item.step for item in self.true_data.notes]
		test_notes = [item.step for item in self.test_data.notes]
		notes_anw = func(true_notes, test_notes)

		true_note_objects = self._rebuild(notes_anw.aligned_true_data, self.true_data.notes)
		test_note_objects = self._rebuild(notes_anw.aligned_test_data, self.test_data.notes)
		for index, objects in enumerate(true_note_objects):
			self._compare(objects, test_note_objects[index])

		# Rests
		true_rests = [return_char_except(item.measure) for item in self.true_data.rests]
		test_rests = [return_char_except(item.measure) for item in self.test_data.rests]
		rests_anw = func(true_rests, test_rests)

		true_rest_objects = self._rebuild(rests_anw.aligned_true_data, self.true_data.rests)
		test_rest_objects = self._rebuild(rests_anw.aligned_test_data, self.test_data.rests)
		for index, objects in enumerate(true_rest_objects):
			self._compare(objects, test_rest_objects[index])

		# Time Signature
		true_timeSignatures = [return_char_except(item.measure) for item in self.true_data.timeSignatures]
		test_timeSignatures = [return_char_except(item.measure) for item in self.test_data.timeSignatures]
		timeSignatures_anw = func(true_timeSignatures, test_timeSignatures)

		true_timeSignature_objects = self._rebuild(timeSignatures_anw.aligned_true_data, self.true_data.timeSignatures)
		test_timeSignature_objects = self._rebuild(timeSignatures_anw.aligned_test_data, self.test_data.timeSignatures)
		for index, objects in enumerate(true_timeSignature_objects):
			self._compare(objects, test_timeSignature_objects[index])

		# Key Signature
		true_keySignatures = [return_char_except(item.measure) for item in self.true_data.keySignatures]
		test_keySignatures = [return_char_except(item.measure) for item in self.test_data.keySignatures]
		keySignatures_anw = func(true_keySignatures, test_keySignatures)

		true_keySignature_objects = self._rebuild(keySignatures_anw.aligned_true_data, self.true_data.keySignatures)
		test_keySignature_objects = self._rebuild(keySignatures_anw.aligned_test_data, self.test_data.keySignatures)
		for index, objects in enumerate(true_keySignature_objects):
			self._compare(objects, test_keySignature_objects[index])

		# Clefs
		true_clefs = [return_char_except(item.measure) for item in self.true_data.clefs]
		test_clefs = [return_char_except(item.measure) for item in self.test_data.clefs]
		clef_anw = func(true_clefs, test_clefs)

		true_clef_objects = self._rebuild(clef_anw.aligned_true_data, self.true_data.clefs)
		test_clef_objects = self._rebuild(clef_anw.aligned_test_data, self.test_data.clefs)
		for index, objects in enumerate(true_clef_objects):
			self._compare(objects, test_clef_objects[index])

		# Spanners
		true_spanners = [item.name for item in self.true_data.spanners]
		test_spanners = [item.name for item in self.test_data.spanners]
		spanner_anw = func(true_spanners, test_spanners)

		true_spanner_objects = self._rebuild(spanner_anw.aligned_true_data, self.true_data.spanners)
		test_spanner_objects = self._rebuild(spanner_anw.aligned_test_data, self.true_data.spanners)
		for index, objects in enumerate(true_spanner_objects):
			self._compare(objects, test_spanner_objects[index])

	def sequence_alignment(self, func):
		"""
		Align each note object with each other based on a scoring method. It
		serves more as a proof-of-concept or a placeholder to be built later.

		.. note:: Align each note object with each other based on the parameters of each.
			These are defined in the scoring_method, for an example look

		:param [func]: Takes a function to be used in the alignment process
		:type [func]: SequenceAlignment
		"""

		# Notes
		true_notes = [item for item in self.true_data.notes]
		test_notes = [item for item in self.test_data.notes]
		notes_anw = func(true_notes, test_notes)

		true_note_objects = self._rebuild(notes_anw.aligned_true_data, self.true_data.notes)
		test_note_objects = self._rebuild(notes_anw.aligned_test_data, self.test_data.notes)
		for index, objects in enumerate(true_note_objects):
			self._compare(objects, test_note_objects[index])

		# Rests
		true_rests = [item for item in self.true_data.rests]
		test_rests = [item for item in self.test_data.rests]
		rests_anw = func(true_rests, test_rests)

		true_rest_objects = self._rebuild(rests_anw.aligned_true_data, self.true_data.rests)
		test_rest_objects = self._rebuild(rests_anw.aligned_test_data, self.test_data.rests)
		for index, objects in enumerate(true_rest_objects):
			self._compare(objects, test_rest_objects[index])

		# Time Signature
		true_timeSignatures = [item for item in self.true_data.timeSignatures]
		test_timeSignatures = [item for item in self.test_data.timeSignatures]
		timeSignatures_anw = func(true_timeSignatures, test_timeSignatures)

		true_timeSignature_objects = self._rebuild(timeSignatures_anw.aligned_true_data, self.true_data.timeSignatures)
		test_timeSignature_objects = self._rebuild(timeSignatures_anw.aligned_test_data, self.test_data.timeSignatures)
		for index, objects in enumerate(true_timeSignature_objects):
			self._compare(objects, test_timeSignature_objects[index])

		# Key Signature
		true_keySignatures = [item for item in self.true_data.keySignatures]
		test_keySignatures = [item for item in self.test_data.keySignatures]
		keySignatures_anw = func(true_keySignatures, test_keySignatures)

		true_keySignature_objects = self._rebuild(keySignatures_anw.aligned_true_data, self.true_data.keySignatures)
		test_keySignature_objects = self._rebuild(keySignatures_anw.aligned_test_data, self.test_data.keySignatures)
		for index, objects in enumerate(true_keySignature_objects):
			self._compare(objects, test_keySignature_objects[index])

		# Clefs
		true_clefs = [item for item in self.true_data.clefs]
		test_clefs = [item for item in self.test_data.clefs]
		clef_anw = func(true_clefs, test_clefs)

		true_clef_objects = self._rebuild(clef_anw.aligned_true_data, self.true_data.clefs)
		test_clef_objects = self._rebuild(clef_anw.aligned_test_data, self.test_data.clefs)
		for index, objects in enumerate(true_clef_objects):
			self._compare(objects, test_clef_objects[index])



# -*- coding: utf-8 -*-
"""
This project wraps around `Music21` and `lxml` to make certain tasks easier and
provide code for lining up and evaluating multiple symbolic music-files. I
do not agree with certain design decisions `Music21` has taken regarding the
handling of the MusicXML format. Issues which become observable to me when
examining the output of multiple music engraving software.

*************
Documentation
*************
"""

from io import BytesIO

from lxml import etree

from mupix.extra import __return_root_path
from mupix.sequence_alignment import (
	AffineNeedlemanWunsch,
	AdvancedAffineNeedlemanWunsch
)
from mupix.typewise import BaseCompareClass
from mupix.partwise import PartiwiseCompareClass


class BasicCompare(BaseCompareClass):
	"""
	Using 1-to-1 comparisons based on the index of each element.
	Obviously not ideal, but can be an interesting comparison.
	"""
	def __init__(self, true_filepath: str, test_filepath: str, do_not_count: list = []):
		super().__init__(true_filepath, test_filepath, do_not_count)
		self._object_split()
		self._total()


class SimpleNeedlemanWunsch(BaseCompareClass):
	"""
	This version aligns notes based on the note names, and the measures + offsets
	of other elements.

	.. note:: Align each note object with each other. Understandably this is not
		the best, but it serves more as a proof-of-concept or a placeholder to be
		built later.

		- Notes           are aligned by step names

		- Restes          are aligned by measure number as a single char

		- TimeSignatures  are aligned by measure number as a single char

		- KeySignatures   are aligned by measure number as a single char

		- Clefs           are aligned by measure number as a single char
	"""
	def __init__(self, true_filepath: str, test_filepath: str, do_not_count: list = []):
		super().__init__(true_filepath, test_filepath, do_not_count)
		self.basic_sequence_alignment(func=AffineNeedlemanWunsch)
		self._total()


class WeightedNeedlemanWunsch(BaseCompareClass):
	"""
	Using a weighted version of Affine Needleman-Wunsch, the way it should be
	used.
	"""
	def __init__(self, true_filepath: str, test_filepath: str, do_not_count: list = []):
		super().__init__(true_filepath, test_filepath, do_not_count)
		self.sequence_alignment(func=AdvancedAffineNeedlemanWunsch)
		self._total()


class PartwiseWeightedNeedlemanWunsch(PartiwiseCompareClass):
	"""
	"""
	def __init__(self, true_filepath: str, test_filepath: str, do_not_count: list = []):
		super().__init__(true_filepath, test_filepath, do_not_count)
		self.sequence_alignment(func=AdvancedAffineNeedlemanWunsch)

	# 	# Parse both files
	# 	self.true_data = MupixObject.partwise_from_filepath(true_filepath)
	# 	self.test_data = MupixObject.partwise_from_filepath(test_filepath)
	# 	self.partwise_sequence_alignment(func=AdvancedAffineNeedlemanWunsch)

	# def partwise_sequence_alignment(self, func):
	# 	for part, _ in enumerate(self.true_data.notes):

	# 		true_notes_part_x = [item for item in self.true_data.notes[part]]
	# 		test_notes_part_x = [item for item in self.test_data.notes[part]]
	# 		notes_x = func(true_notes_part_x, test_notes_part_x)  # noqa

	# 		try:
	# 			true_notes_part_y = [item for item in self.true_data.notes[part]]
	# 			test_notes_part_y = [item for item in self.test_data.notes[part + 1]]
	# 			notes_y = func(true_notes_part_y, test_notes_part_y)
	# 		except IndexError:
	# 			notes_y = float("inf")  # noqa


def xml_validator(musicxml_filepath, schema_filepath=__return_root_path() + "/tests/xml/musicxml.xsd"):
	"""Return if the provided musicxml file is valid against the current musicxml schema.

	:param [musicxml_filepath]: A character string that represents the filepath and filename of the file to open.
	:type [musicxml_filepath]: String

	:param [schema_filepath](optional): A character string that represents the filepath and filename of the schema to use.
	:type [schema_filepath]: String

	:return: Returns a boolean value, either it is a valid MusicXML file or it is not.
	:rtype: Bool
	"""
	with open(musicxml_filepath, "rb") as xml_file:
		test = BytesIO(xml_file.read())

	try:
		xml_schema = etree.XMLSchema(etree.parse(schema_filepath))
	except etree.XMLSyntaxError:
		xml_schema = etree.DTD(schema_filepath)

	return xml_schema.validate(etree.parse(test))


def xml_type_finder(musicxml_filepath):
	"""
	Check if the xml file is written in a partwise or timewise fashion. This is just the fastest way
	to check.

	:param [filepath]: A character string that represents the filepath and filename of the file to open.
	:type [filepath]: String

	:return: Either a Partwise or a Timewise string will return.
	:rtype: String
	"""
	with open(musicxml_filepath, "r") as xml_file:
		for line in xml_file:
			if "score-partwise" in line.lower():
				return "Partwise"
			elif "score-timewise" in line.lower():
				return "Timewise"
			else:
				pass  # Leave room 
		raise Exception("File has neither time-wise or part-wise tags")



import copy
import json
import os
import re
import sys

from music21.interval import Interval
from music21.key import Key
from music21.note import Note


def __return_root_path() -> str:
	"""
	This is a simple function to get the root directory of the repository.

	Returns:
		(string): of the local filepath to this project.
	"""
	if os.name == "posix":
		return "/".join(os.path.abspath(__file__).split("/")[0:-2])
	else:
		return "\\".join(os.path.abspath(__file__).split("\\")[0:-2])


def return_char_except(value: int) -> str:
	"""
	Because we are already using the "_" to signal a skipped Mupix Object.
	"""
	if value >= 95:
		value += 1
	return chr(value)


def boundary_search(begin, end, string):
	"""
	Boundary Search
	This is the regex formula for finding everything between two specific strings.
	e.g. :
				this character =>[ (all of the stuff here, including newlines) ] <= this character
				(?<=[)(.*?)(?=])
	Args:
		begin (string): The character that signals the begins what you want to search for.
		end (string): The character that signals the end of what you are looking for.
		string (string): Multi-line data you wish to search through.
	Returns:
		output (list): A list of all instances of what was searched for.
	"""
	pattern = f"(?<={begin})(.*?)(?={end})"
	return re.findall(pattern, string, re.DOTALL)


def output_filter(ctx, func, *args, **kwargs):
	"""
	This is a helper function to filter the output of all commands into notes, rests, etc.
	"""
	output = func(*args, **kwargs)
	msg = {}

	# If no filtering options are defined, output all information
	if all(value == False for value in [ctx["notes"], ctx["rests"], ctx["time_signatures"], ctx["key_signatures"], ctx["clefs"], ctx["spanners"], ctx["dynamics"], ctx["error_description"]]):  # noqa
		# turn to json serializable.
		msg["Notes"] = [i.asdict() for i in output.notes]
		msg["Rests"] = [i.asdict() for i in output.rests]
		msg["TimeSignatures"] = [i.asdict() for i in output.timeSignatures]
		msg["KeySignatures"] = [i.asdict() for i in output.keySignatures]
		msg["Clefs"] = [i.asdict() for i in output.clefs]
		msg["Spanners"] = [i.asdict() for i in output.spanners]
		msg["Dynamics"] = [i.asdict() for i in output.dynamics]
		# Leave out of the regular output
		# msg["ErrorDescription"] = output.error_description
	else:
		# Not Pretty Print with a combination
		if ctx["notes"]:
			msg["Notes"] = [i.asdict() for i in output.notes]
		if ctx["rests"]:
			msg["Rests"] = [i.asdict() for i in output.rests]
		if ctx["time_signatures"]:
			msg["TimeSignatures"] = [i.asdict() for i in output.timeSignatures]
		if ctx["key_signatures"]:
			msg["KeySignatures"] = [i.asdict() for i in output.keySignatures]
		if ctx["clefs"]:
			msg["Clefs"] = [i.asdict() for i in output.clefs]
		if ctx["spanners"]:
			msg["Spanners"] = [i.asdict() for i in output.spanners]
		if ctx["dynamics"]:
			msg["Dynamics"] = [i.asdict() for i in output.dynamics]
		if ctx["error_description"]:
			msg["ErrorDescription"] = output.error_description

	if ctx["total_only"]:
		for category in msg:
			try:
				msg[category] = [item for item in msg[category] if 'Total' in item['name']]
			except KeyError:
				print(f"[{func}] - Does not have a Total output or something went wrong with it.")
				sys.exit(0)

	if not ctx["pretty_print"]:
		print(msg)
	else:
		print(json.dumps(msg, indent=2))


def _populate_list(input_list, maximum):
	"""
	Different music engraving software encode time signatures, key signatures,
	and clefs in a peculiar way. Some repeat the previous object in every
	following measure.

	Let's say this hypothetical score has 44 measures. When parsing a problematic musicXML file,
	you will notice there are 44 treble clefs in the same instrument part. Meaning the same,
	unchanging clef, was repeated once per measure in the file. The same can happen on a per-system
	basis. It is not a bad way to explain the music because each element implied, even if it is missing.

	To avoid any further complication, the clefs, time signatures, and key signatures have been
	"normalized" to always be repeated on every measure, because they are still "acting" on a measure when
	they are omitted.

	[TODO] Clean up the algorithm, maybe splitting it into an iter and a yielder or something with dequeue
	for better readability. This is just a PoC anyway.

	:param [input_list]: A list of objects (KeySignatureObject, TimeSignatureObject or ClefObject).
	:type [input_list]: List

	:param [maximum]: An integer that represents the maximum number of measures in the score. In other words, how many measures should the objects be expanded for.
	:type [maximum]: Integer

	:return: Returns a the original list with added objects if they are missing from each measure.
	:rtype: List
	"""

	measure = 1
	output_list = []
	# If there are no inputs given, skip everything.
	if len(input_list) == 0:
		return []

	while measure <= maximum:
		try:
			# Next element follows sequential number
			if input_list[0].measure == measure:
				output_list.append(input_list.pop(0))
				measure += 1
			# Missing element, create a new one and add it at the end of the list
			elif input_list[0].measure > measure:
				# If there is no information in the first measure, pick the information from the first element and duplicate
				if len(output_list) == 0:
					last_object = copy.deepcopy(input_list[0])
					last_object.measure = measure
					output_list.append(last_object)
					measure += 1
				else:
					last_object = copy.deepcopy(output_list[-1])
					last_object.measure += 1
					output_list.append(last_object)
					measure += 1
			# Next element has the same number as the previous one, but the other attributes are different.
			elif input_list[0].measure < measure:
				output_list.append(input_list.pop(0))
		# `input_list` last item's number was smaller than maximum.
		# Keep creating a duplicate of the last item until maximum is reached.
		except IndexError:
			try:
				tail_object = copy.deepcopy(output_list[-1])
			except IndexError:
					raise Exception(len(input_list), input_list, output_list)
			tail_object.measure += 1
			output_list.append(tail_object)
			measure += 1
	return output_list


def normalize_object_list(object_list, total_measures, total_parts):
	"""
	Without changing the order of the objects, iterate through each instrumental
	part and populate missing information in place. This small function just
	makes sure :func:`mupix.base._populate_list` occurs on every part.
	"""
	obj = []
	for part in range(1, total_parts + 1):
		obj += _populate_list([x for x in object_list if x.part == part], total_measures)
	return obj


def add_step_information(notes: list, keySignatures: list):
	"""
	This function will populate the step information into Mupix note objects, it
	is required because music21 will not keep key signature information in
	measure other than the measure the key is defined in when reading musicXML.
	The maintainers of music21 don't believe this is an issue and won't fix it,
	so this and others must exist.

	:param [notes]: A list of Mupix NoteObjects.
	:type [notes]: List

	:param [keySignatures]: A list of Mupix KeySignatureObjects.
	:type [keySignatures]: List

	:return [List]: The original list of Mupix NoteObjects (in order) with step information included.
	:rtype: List
	"""
	for key in keySignatures:  # min 1 keySignatures per measure

		key_name = key.step.upper() if key.mode == "major" else key.step.lower()
		for note in notes:
			if note.part == key.part and note.measure == key.measure:
				note.step = Interval(noteStart=Note(Key(key_name).asKey().tonic), noteEnd=Note(note._music21_object.step)).semitones % 12

	return notes


def _temp_fix_spanners(spanner: list, tmp_list: list) -> list:
	"""
	"""
	if len(spanner) != len(tmp_list):
		raise Exception("Spanner measure addition error.")

	for index, item in enumerate(spanner):
		spanner[index].measure = tmp_list[index]["measure"]

	return spanner



import itertools
import operator

import attr
import music21

from mupix.core import (
	NoteObject,
	RestObject,
	TimeSignatureObject,
	KeySignatureObject,
	ClefObject,
	Marking,
)
from mupix.extra import (
	add_step_information,
	normalize_object_list,
)
from mupix.typewise import BaseCompareClass


@attr.s
class MupixPartwiseObject():
	parts = attr.ib(
		kw_only=True,
		type=list,
		validator=attr.validators.deep_iterable(
			member_validator=attr.validators.instance_of(list),
			iterable_validator=attr.validators.instance_of(list),
		)

		# validator=attr.validators.deep_iterable(
		# 	member_validator=attr.validators.deep_iterable(
		# 		member_validator=Marking,
		# 		iterable_validator=attr.validators.instance_of(list)
		# 	),
		# 	iterable_validator=attr.validators.instance_of(list)
		# ),
	)

	def __iter__(self):
		return iter(self.parts)

	@classmethod
	def from_filepath(cls, filepath):
		"""

		Music notation software will sometimes repeat time signature, key signature
		or clef information arbitrarily. To combat this, we repeat the key, time
		and clef information in all measures because it always acts on the
		information inside of the measure anyway.

		Sometimes key signature information gets deleted in music21 objects, to
		combat this we force the key signature to be reintroduced to specify a note
		step.
		"""

		notes, rests, timeSignatures, keySignatures, clefs = [], [], [], [], []
		# Creates lists of lists notes[part][index]
		for parts_index, parts in enumerate(music21.converter.parseFile(filepath).recurse().getElementsByClass("Part"), 1):  # noqa
			notes.append([NoteObject(item, parts_index) for item in parts.recurse().notes if not item.isChord])
			rests.append([RestObject(item, parts_index) for item in parts.recurse().notesAndRests if not item.isNote])
			timeSignatures.append([TimeSignatureObject(item, parts_index) for item in parts.recurse().getElementsByClass("TimeSignature")])  # noqa
			keySignatures.append([KeySignatureObject(item, parts_index) for item in parts.recurse().getElementsByClass("KeySignature")])  # noqa
			clefs.append([ClefObject(item, parts_index) for item in parts.recurse().getElementsByClass("Clef")])

		try:
			measuresInScore = max(notes[0] + rests[0], key=operator.attrgetter('measure')).measure
		except ValueError:
			measuresInScore = 0
			parts_index = 0

		# Make sure the correct key and measure information is in every measure.
		# Sometimes
		for part in range(parts_index):
			timeSignatures[part] = normalize_object_list(
				object_list=timeSignatures[part],
				total_measures=measuresInScore,
				total_parts=parts_index
			)
			keySignatures[part] = normalize_object_list(
				object_list=keySignatures[part],
				total_measures=measuresInScore,
				total_parts=parts_index
			)
			clefs[part] = normalize_object_list(
				object_list=clefs[part],
				total_measures=measuresInScore,
				total_parts=parts_index
			)

		# Make sure the key signature information is in the note objects
		for part in range(parts_index):
			# only once keySignatures are normalized can we add the step information.
			notes[part] = add_step_information(notes[part], keySignatures[part])

		# Finally sort all elements in the correct musical stream
		out = []
		for part in range(parts_index):
			# Get notes and rests together first
			notes_and_rests = sorted(
				itertools.chain(notes[part], rests[part]),
				key=operator.attrgetter("measure", "onset")
			)
			# Make sure clefs, time sigs, and key sigs come first.
			out.append(
				sorted(
					itertools.chain(
						clefs[part],
						timeSignatures[part],
						keySignatures[part],
						notes_and_rests,
					),
					key=operator.attrgetter("measure", "onset")
				)
			)
		return cls(parts=out)


class PartiwiseCompareClass(BaseCompareClass):
	"""
	"""
	def __init__(self, true_filepath: str, test_filepath: str, do_not_count: list = []):
		self.true_data = MupixPartwiseObject.from_filepath(true_filepath)
		self.test_data = MupixPartwiseObject.from_filepath(test_filepath)

	def sequence_alignment(self, func):

		for part in self.true_data:
			true_part = part
			break
		for part in self.test_data:
			test_part = part
			break

		# print(true_part)
		a = func(true_part, test_part)
		print(a)
		import sys
		sys.exit(0)
